/*
 * info.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Dec 2008 - Created.
 */

/* 
 * info.c:
 * Refered to Wine code
 */

#include <linux/fs.h>
#include "win32_process.h"
#include "wineserver/info.h"
#include "wineserver/uk_lib.h"
#include "wineserver/file.h"

#ifdef CONFIG_UNIFIED_KERNEL
#define IS_HANDLE_VALID(handle)   (((ULONG_PTR)handle & UK_HANDLE_FLAG) \
		&& !((ULONG_PTR)handle & KERNEL_HANDLE_FLAG))

struct fd;
struct file *get_unix_file(struct fd *fd);

static inline unsigned long get_handle_index(obj_handle_t handle)
{
	if (!IS_HANDLE_VALID(handle))
		return 0;

	return ((unsigned long)UK_HANDLE_TO_HANDLE(handle) >> 2);
}

struct handle_info_table *alloc_handle_info_table(void)
{
	struct handle_info_table *table;

	table = (struct handle_info_table *)kmalloc(sizeof(struct handle_info_table), GFP_KERNEL);
	if (!table)
		return NULL;

	table->used = 0;
	table->order = 0;
	table->allocated = PAGE_SIZE;
	table->handles = (struct handle_info *)__get_free_page(GFP_KERNEL);
	if (!table->handles) {
		kfree(table);
		return NULL;
	}

	memset(table->handles, 0, PAGE_SIZE);

	return table;
}

int expand_handle_info_table(struct handle_info_table *table, unsigned long new_size)
{
	int order;
	unsigned long new_alloc;
	struct handle_info *new_handles;

	new_alloc = table->allocated;
	order = table->order;
	while (new_alloc < new_size) {
		new_alloc <<= 1;
		order++;
	}

	new_handles = (struct handle_info *)__get_free_pages(GFP_KERNEL, order);
	if (!new_handles)
		return -ENOMEM;

	memcpy(new_handles, table->handles, table->allocated);
	free_pages((unsigned long)table->handles, table->order);

	table->used = new_size;
	table->order = order;
	table->allocated = new_alloc;
	table->handles = new_handles;

	return 0;
}

void free_handle_info_table(struct handle_info_table *table)
{
	free_pages((unsigned long)table->handles, table->order);
}

int set_handle_info(struct eprocess *process, obj_handle_t handle, void *obj)
{
	int unix_fd;
	struct handle_info *info;
	struct fd *fd;
	struct file *unix_file;
	unsigned long index;
	struct handle_info_table *table;

	if (!process || !process->ep_handle_info_table)
		return -EINVAL;
	table = process->ep_handle_info_table;

	index = get_handle_index(handle);
	if (!index)
		return -EINVAL;

	if ((((index + 1) << 3) > table->allocated)
			&& expand_handle_info_table(table, (index + 1) << 3))
		return -ENOMEM;

	info = &table->handles[index];

	fd = get_obj_fd(obj);
	if (fd) {
		unix_fd = get_unused_fd();
		if (unix_fd < 0) {
			release_object(fd);
			return unix_fd;
		}

		unix_file = get_unix_file(fd);
		get_file(unix_file);
		fd_install(unix_fd, unix_file);

		release_object(fd);

		info->unix_fd = unix_fd + 1;
		info->obj = obj;
	}

	if (table->used <= (index << 3))
		table->used += sizeof(struct handle_info);

	return 0;
}

struct handle_info *get_handle_info(struct eprocess *process, obj_handle_t handle)
{
	unsigned long index;
	struct handle_info_table *table;

	if (!process || !process->ep_handle_info_table)
		return NULL;

	index = get_handle_index(handle);
	table = process->ep_handle_info_table;

	return index ? &table->handles[index] : NULL;
}

void clear_handle_info(struct eprocess *process, obj_handle_t handle)
{
	unsigned long index;
	struct handle_info_table *table;
	struct handle_info *info;

	if (!process || !process->ep_handle_info_table)
		return;

	index = get_handle_index(handle);
	if (!index)
		return;

	table = process->ep_handle_info_table;
	info = &table->handles[index];

	if (info->unix_fd)
		uk_close(info->unix_fd - 1);
	memset(info, 0, sizeof(struct handle_info));

	if (((index + 1) << 3) >= table->used)
		table->used -= sizeof(struct handle_info);
}

int get_handle_fd(struct eprocess *process, obj_handle_t handle)
{
	struct handle_info *info = get_handle_info(process, handle);

	return info ? info->unix_fd - 1 : -1;
}
#endif
